//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/main.rs: Main entry point
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! syd: seccomp and landlock based application sandbox with support for namespaces
//! Main entry point.

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    collections::HashSet,
    env,
    ffi::{CString, OsStr},
    fs::File,
    io::Write,
    os::unix::{ffi::OsStrExt, process::CommandExt},
    path::{Path, PathBuf},
    process::{exit, Command, ExitCode},
    str::FromStr,
};

use anyhow::{bail, Context, Error};
use getargs::{Opt, Options};
use libseccomp::{ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall, ScmpVersion};
use nix::{
    errno::Errno,
    libc::{setdomainname, STDERR_FILENO},
    mount::{mount, umount2, MntFlags, MsFlags},
    sched::{unshare, CloneFlags},
    sys::{
        resource::{setrlimit, Resource},
        signal::{signal, SigHandler, Signal},
        utsname::uname,
        wait::{waitpid, WaitStatus},
    },
    unistd::{chdir, chroot, fork, getgid, getuid, pivot_root, sethostname, write, ForkResult},
    NixPath,
};
use once_cell::sync::Lazy;
use parse_size::parse_size;
use syd::{
    bring_up_loopback,
    config::*,
    error,
    hook::{ExportMode, Supervisor},
    info,
    landlock::{
        path_beneath_rules, Access, AccessFs, RestrictionStatus, Ruleset, RulesetAttr,
        RulesetCreatedAttr, RulesetError, RulesetStatus, ABI,
    },
    log::JsonLinesLogger,
    proc::proc_limit_userns,
    sandbox::Sandbox,
    unshare::{GidMap, UidMap},
};

static SYD_VERSION: Lazy<&'static str> = Lazy::new(|| {
    if env!("SYD_GITHEAD").is_empty() {
        env!("CARGO_PKG_VERSION")
    } else {
        env!("SYD_GITHEAD")
    }
});

// This line will include the content of "esyd.sh" at compile-time
const ESYD: &str = include_str!("esyd.sh");

// Landlock API in use.
const LANDLOCK_ABI: ABI = ABI::V3;
const LANDLOCK_ABI_STR: &str = "v3";

const NONE: Option<&PathBuf> = None::<PathBuf>.as_ref();

#[derive(Debug, Default)]
struct Arguments {
    root: Option<PathBuf>,
    map_root: bool,

    arg0: Option<String>,
    argv: Vec<String>,

    export: Option<ExportMode>,

    sandbox: Sandbox,
}

// A helper function to wrap the operations and reduce duplication
fn landlock_operation(
    path_ro: &[String],
    path_rw: &[String],
    abi: ABI,
) -> Result<RestrictionStatus, RulesetError> {
    let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
    let created_ruleset = ruleset.create()?;
    let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
    let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
    let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
    let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
    final_ruleset.restrict_self().map_err(RulesetError::from)
}

// Set SIGPIPE handler to default.
fn setup_sigpipe() -> Result<(), Errno> {
    // SAFETY: The nix::sys::signal::signal function is unsafe because
    // it affects the global state of the program by changing how a
    // signal (SIGPIPE in this case) is handled. It's safe to call here
    // because changing the SIGPIPE signal to its default behavior will
    // not interfere with any other part of this program that could be
    // relying on a custom SIGPIPE signal handler.
    unsafe { signal(Signal::SIGPIPE, SigHandler::SigDfl) }.map(|_| ())
}

#[allow(clippy::cognitive_complexity)]
fn main() -> anyhow::Result<ExitCode> {
    // Set SIGPIPE handler to default.
    setup_sigpipe().context("Failed to set SIGPIPE signal handler to default.")?;

    // Initialize logging.
    JsonLinesLogger::init().context("Failed to initialize logging.")?;

    // Parse CLI arguments
    let (argv0, args) = parse_argv().context("Failed to parse command line arguments")?;

    // Handle namespace re-exec
    const ENV_INIT: &str = "SYD_INIT";
    let init = env::var(ENV_INIT).is_err();
    env::remove_var(ENV_INIT);
    if !init {
        return run(&argv0, args);
    }

    // Limit process resources.
    let restrict_prlimit = !args.sandbox.allow_unsafe_prlimit();
    if restrict_prlimit {
        if let Err(errno) = setrlimit(Resource::RLIMIT_CORE, 0, 0) {
            error!("ctx": "limit_core", "err": errno.to_string());
        }
    }

    // Set up Linux namespaces if requested. Note,
    // we set it up here before spawning the child so as to
    // include the SydB☮x process into the pid namespace as well
    // such that the sandbox process and SydB☮x have the identical
    // view of /proc.
    let mut namespaces = CloneFlags::empty();
    for namespace in args.sandbox.namespaces() {
        namespaces |= syd::unshare::namespace::to_clone_flag(namespace);
    }
    if namespaces.contains(CloneFlags::CLONE_NEWPID) {
        // SAFETY: We cannot support NEWPID without NEWNS.
        // ie, pid namespace must have its own private /proc.
        namespaces.insert(CloneFlags::CLONE_NEWNS);
    }
    if namespaces.is_empty() {
        // No namespace arguments passed, run normally.
        return run(&argv0, args);
    }

    // Prepare command to reexec SydB☮x asserting initialization is done.
    let exe = env::current_exe().context("Failed to get current executable")?;
    let arg: Vec<String> = env::args().skip(1).collect();
    let mut syd = std::process::Command::new(exe);
    syd.args(&arg);
    syd.env(ENV_INIT, "1");

    // create the UID and GID mappings.
    let uid = getuid().as_raw();
    let gid = getgid().as_raw();

    let id_buf = if args.sandbox.unshare_user() {
        let uid_buf = {
            let uid_maps = vec![
                UidMap {
                    inside_uid: if args.map_root { 0 } else { uid },
                    outside_uid: uid,
                    count: 1,
                }, // Map the current user.
            ];
            let mut buf = Vec::new();
            for map in uid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_uid, map.outside_uid, map.count
                )
                .context("Failed to format uidmap")?;
            }
            buf
        };

        let gid_buf = {
            let gid_maps = vec![
                GidMap {
                    inside_gid: if args.map_root { 0 } else { gid },
                    outside_gid: gid,
                    count: 1,
                }, // Map the current group.
            ];
            let mut buf = Vec::new();
            for map in gid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_gid, map.outside_gid, map.count
                )
                .context("Failed to format gidmap")?;
            }
            buf
        };
        Some((uid_buf, gid_buf))
    } else {
        None
    };

    // We need this to make --map-root work in user subnamespace.
    let id_buf0 = if args.sandbox.unshare_user() {
        let uid_buf = {
            let uid_maps = vec![
                UidMap {
                    inside_uid: if args.map_root { 0 } else { uid },
                    outside_uid: if args.map_root { 0 } else { uid },
                    count: 1,
                }, // Map the current user.
            ];
            let mut buf = Vec::new();
            for map in uid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_uid, map.outside_uid, map.count
                )
                .context("Failed to format uidmap")?;
            }
            buf
        };

        let gid_buf = {
            let gid_maps = vec![
                GidMap {
                    inside_gid: if args.map_root { 0 } else { gid },
                    outside_gid: if args.map_root { 0 } else { gid },
                    count: 1,
                }, // Map the current group.
            ];
            let mut buf = Vec::new();
            for map in gid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_gid, map.outside_gid, map.count
                )
                .context("Failed to format gidmap")?;
            }
            buf
        };
        Some((uid_buf, gid_buf))
    } else {
        None
    };

    // Create namespace.
    unshare(namespaces).context("Failed to create namespace")?;

    // SAFETY: If entering into a new user namespace:
    // 1. Limit max_user_namespaces to 1.
    // 2. Enter into a second user namespace.
    // This way we prevent sandbox process from creating subnamespaces
    // with a possibly different view of the filesystem. This is
    // functionally identical to bubblewrap's --disable-userns option.
    // We must to this by default unconditionally because SydB☮x must
    // have an identical view of the filesystem with the sandbox process
    // to operate correctly.
    if namespaces.contains(CloneFlags::CLONE_NEWUSER) {
        proc_limit_userns().context("Failed to limit maximum user namespaces")?;
    }

    // Write uid/gid map for user namespace.
    if let Some((ref uid_buf, ref gid_buf)) = id_buf {
        // Write "deny" to /proc/self/setgroups before writing to gid_map.
        File::create("/proc/self/setgroups")
            .and_then(|mut f| f.write_all(b"deny"))
            .context("Failed to write to /proc/self/setgroups")?;
        File::create("/proc/self/gid_map")
            .and_then(|mut f| f.write_all(&gid_buf[..]))
            .context("Failed to write to /proc/self/gid_map")?;
        File::create("/proc/self/uid_map")
            .and_then(|mut f| f.write_all(&uid_buf[..]))
            .context("Failed to write to /proc/self/uid_map")?;
    }

    // Bring up loopback device for net namespace.
    if args.sandbox.unshare_net() {
        // Set up the loopback interface.
        if let Err(errno) = bring_up_loopback() {
            error!("ctx": "ifup_lo", "err": errno.to_string());
        }
    }

    // Set host and domain name for uts namespace.
    if args.sandbox.unshare_uts() {
        // SAFETY: API_VERSION must not contain a nul-byte.
        #[allow(clippy::disallowed_methods)]
        let domainname = CString::new(API_VERSION.to_string()).unwrap();

        // SAFETY: There's no setdomainname wrapper in nix.
        let _ = unsafe { setdomainname(domainname.as_ptr() as *const _, domainname.len()) };
        let _ = sethostname("SydB☮x");
    }

    // SAFETY: In libc we trust.
    match unsafe { fork() }.context("Failed to fork")? {
        ForkResult::Parent { child, .. } => {
            // SAFETY: Set up a Landlock sandbox to disallow all access.
            let _ = landlock_operation(&[], &[], LANDLOCK_ABI);

            // SAFETY: Set up a seccomp filter which only allows
            // 1. read, write and exit
            // 2. wait4 and waitpid
            // 3. memory allocation functions
            let mut filter = ScmpFilterContext::new_filter(ScmpAction::KillProcess)
                .context("Failed to create seccomp filter")?;
            let allow_call = [
                "read",
                "write",
                "exit",
                "exit_group",
                "wait4",
                "waitpid",
                "brk",
                "madvise",
                "mmap",
                "mmap2",
                "munmap",
                "sigaltstack",
            ];
            for call in allow_call {
                filter.add_rule(ScmpAction::Allow, ScmpSyscall::new(call))?;
            }
            filter.load().context("Failed to load seccomp filter")?;

            loop {
                match waitpid(child, None) {
                    Ok(WaitStatus::Exited(_, code)) => {
                        #[allow(clippy::cast_possible_truncation)]
                        #[allow(clippy::cast_sign_loss)]
                        return Ok(ExitCode::from(code as u8));
                    }
                    Ok(WaitStatus::StillAlive) | Err(Errno::EINTR) => {}
                    Ok(_) => {
                        return Ok(ExitCode::from(127));
                    }
                    Err(error) => {
                        bail!("Failed to wait: {error}");
                    }
                }
            }
        }
        ForkResult::Child => {
            // SAFETY: It is unsafe to allocate here.
            // println!, unwrap() etc. are also unsafe.

            // Set mount propagation on the root filesystem for mount namespace.
            if args.sandbox.unshare_mount() {
                const NONE: Option<&PathBuf> = None::<PathBuf>.as_ref();
                if let Some(flags) = args.sandbox.propagation {
                    if !flags.is_empty() {
                        if let Err(errno) = mount(Some("none"), "/", NONE, flags, NONE) {
                            write(
                                STDERR_FILENO,
                                b"Failed to change root filesystem propagation\n",
                            )
                            .ok();
                            // SAFETY: _exit is async-signal-safe.
                            unsafe { nix::libc::_exit(errno as i32) };
                        }
                    }
                }

                if let Some(ref proc) = args.sandbox.proc {
                    let flags = MsFlags::MS_NOSUID | MsFlags::MS_NOEXEC | MsFlags::MS_NODEV;
                    if let Err(errno) = mount(Some("proc"), proc, Some("proc"), flags, NONE) {
                        write(STDERR_FILENO, b"Failed to mount proc filesystem\n").ok();
                        // SAFETY: _exit is async-signal-safe.
                        unsafe { nix::libc::_exit(errno as i32) };
                    }
                }
            }

            // SAFETY: Enter into a user subnamespace to prevent
            // further namespace modification.
            if namespaces.contains(CloneFlags::CLONE_NEWUSER) {
                if let Err(errno) = unshare(CloneFlags::CLONE_NEWUSER) {
                    write(STDERR_FILENO, b"Failed to enter into user subnamespace\n").ok();
                    // SAFETY: _exit is async-signal-safe.
                    unsafe { nix::libc::_exit(errno as i32) };
                }

                // Write uid/gid map for user subnamespace.
                // Note we use `id_buf0` here rather than `id_buf`
                // which honours --map-root for subnamespace.
                if let Some((ref uid_buf, ref gid_buf)) = id_buf0 {
                    // Write "deny" to /proc/self/setgroups before writing to gid_map.
                    if File::create("/proc/self/setgroups")
                        .and_then(|mut f| f.write_all(b"deny"))
                        .is_err()
                    {
                        write(STDERR_FILENO, b"Failed to write to /proc/self/setgroups\n")?;
                        // SAFETY: _exit is async-signal-safe.
                        unsafe { nix::libc::_exit(nix::libc::EIO) };
                    }
                    if File::create("/proc/self/gid_map")
                        .and_then(|mut f| f.write_all(&gid_buf[..]))
                        .is_err()
                    {
                        write(STDERR_FILENO, b"Failed to write to /proc/self/gid_map\n")?;
                        // SAFETY: _exit is async-signal-safe.
                        unsafe { nix::libc::_exit(nix::libc::EIO) };
                    }
                    if File::create("/proc/self/uid_map")
                        .and_then(|mut f| f.write_all(&uid_buf[..]))
                        .is_err()
                    {
                        write(STDERR_FILENO, b"Failed to write to /proc/self/uid_map\n")?;
                        // SAFETY: _exit is async-signal-safe.
                        unsafe { nix::libc::_exit(nix::libc::EIO) };
                    }
                }
            }

            // Rexecute SydB☮x in the new namespace.
            syd.exec();
            // SAFETY: _exit is async-signal-safe.
            unsafe { nix::libc::_exit(127) };
        }
    }
}

fn run(argv0: &str, mut args: Arguments) -> anyhow::Result<ExitCode> {
    // Change root directory if requested.
    if let Some(ref root) = args.root {
        let proc = root.join("proc");
        if !proc.is_dir() {
            bail!("Directory proc does not exist under root directory!");
        }
        mount(Some(root), root, NONE, MsFlags::MS_BIND, NONE)
            .context("Failed to bind mount the new root directory")?;

        mount(Some("/proc"), &proc, NONE, MsFlags::MS_BIND, NONE)
            .context("Failed to bind mount the /proc directory")?;

        chdir(root).context("Failed to change to root directory")?;
        pivot_root(".", "./proc").context("Failed to pivot root")?;
        umount2("/proc", MntFlags::MNT_DETACH).context("Failed to unmount old root")?;
        chdir("/").context("Failed to change to new root directory")?;

        chroot(".").context("Failed to change root directory")?;
    }

    // Initialize sandbox environment, drop capabilities.
    Supervisor::init_env()?;

    // Set up the Landlock sandbox if requested. Note,
    // we set it up here before spawning the child so as to
    // include the SydB☮x sandbox threads into the sandbox as
    // well. This is done for added security.
    // Note, Landlock errors are not fatal.
    if let Some((mut path_ro, mut path_rw)) = args.sandbox.collect_landlock() {
        path_ro.sort();
        path_rw.sort();

        match landlock_operation(&path_ro, &path_rw, LANDLOCK_ABI) {
            Ok(status) => {
                info!("ctx": "landlock",
                "abi": LANDLOCK_ABI_STR,
                "path_ro": path_ro,
                "path_rw": path_rw,
                "status": match status.ruleset {
                    // The FullyEnforced case must be tested by the developer.
                    RulesetStatus::FullyEnforced => "fully_enforced",
                    RulesetStatus::PartiallyEnforced => "partially_enforced",
                    // Users should be warned that they are not protected.
                    RulesetStatus::NotEnforced => "not_enforced",
                });
            }
            Err(error) => {
                info!("ctx": "landlock",
                    "abi": LANDLOCK_ABI_STR,
                    "path_ro": path_ro,
                    "path_rw": path_rw,
                    "status": "unsupported",
                    "err": error.to_string());
            }
        }
    }

    // Finalize Sandbox, build GlobSets.
    args.sandbox
        .build_globsets()
        .context("Failed to build GlobSets.")?;

    // Step 6: Initialize sandbox supervisor.
    let wait_all = args.sandbox.exit_wait_all();
    let mut supervisor =
        Supervisor::new(args.sandbox, *syd::NPROC, args.export).context(format!(
            "Error creating sandbox with {} threads and export mode {:?}.",
            *syd::NPROC,
            args.export,
        ))?;
    supervisor.init();

    // Prepare the command to execute.
    let mut command = syd::unshare::Command::new(argv0);
    command.args(args.argv);
    if let Some(arg0) = args.arg0 {
        command.arg0(arg0);
    }

    // Spawn the program under sandbox.
    let (pid, thread_handle, pool) = match supervisor.spawn(command) {
        Ok((pid, thread_handle, pool)) => (pid, thread_handle, pool),
        Err(error) => {
            let errno = Errno::last() as u8;
            let error = error.context("Failed to spawn command under sandbox.");
            eprintln!("{error:?}");
            return Ok(ExitCode::from(errno));
        }
    };

    // Wait for the process to exit and return the same error code.
    let code = match Supervisor::wait(pid, thread_handle, pool, wait_all) {
        Ok(code) => code,
        Err(errno) => {
            let error = Error::new(errno).context("Failed to wait for sandboxed process.");
            eprintln!("{error:?}");
            errno as i32
        }
    };
    Ok(ExitCode::from(
        u8::try_from(code).context("Invalid exit code!")?,
    ))
}

fn help() {
    let libapi = libseccomp::get_api();
    let libver = ScmpVersion::current().ok();
    let (libmaj, libmin, libmic) = libver
        .map(|v| (v.major, v.minor, v.micro))
        .unwrap_or((0, 0, 0));
    let native = syd::seccomp_arch_native_name().unwrap_or("?");

    println!(
        "SydB☮x {} libsecc☮mp v{}.{}.{} api:{} arch:{}
seccomp and landlock based application sandbox with support for namespaces
Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
SPDX-License-Identifier: GPL-3.0-or-later

# Usage

syd [-hvcl]
    [--lock] [--root] [--chroot=root]
    [-m command...] [-C path...] [-p name...]
    [--unshare-mount,uts,ipc,user,pid,net,cgroup]
    [-A arg0] [--map-root]
    [--] {{command [arg...]}}
syd --check
syd --landlock
syd --export bpf|pfc
syd --list allow|deny|hook|ioctl|prctl
syd --print profile-name
syd --errno number|name-regex
syd --parse size|human-size
syd --read path
syd --syscall number|name-regex [--arch list|native|x86|x86_64|aarch64...]
syd --sh
syd --test [<name-regex>|<number>|<number>..<number>]..
syd exec {{command [arg...]}}
syd log

-h            -- Show usage and exit
-v            -- Show version and exit
-c            -- Causes command to be executed under a shell with the \"user\" profile
                 Shell is /bin/bash by default, override with SYD_SH
                 Login shell compatibility
-l            -- Ignored
                 Login shell compatibility
--lock        -- Lock magic commands
--root        -- Return 0 as user/group id, aka \"fakeroot\" mode
-m command    -- Run a magic command during init, may be repeated
-C path       -- Path to the configuration file to use, may be repeated
-p name       -- Name of the sandbox profile to use, may be repeated

--check       -- Exit with success if the process is running under SydB☮x
--landlock    -- Exit with success if LandLock ABI v3 is fully supported
--export mode -- Export seccomp rules in given format to standard output and exit
--list set    -- Print the names of the system calls which belong to the given set and exit
                 If set is \"ioctl\", print the list of allowlisted ioctl requests
                 If set is \"prctl\", print the list of allowlisted prctl options
--print name  -- Print out the rules of the given sandbox profile and exit
--errno arg   -- Given a number, print the matching errno name and exit
                 Given a regex, print case-insensitively matching errno names and exit
--parse size  -- Given a number, print human-formatted size and exit
                 Given a string, parse human-formatted size into bytes, print and exit
--read path   -- Print the canonicalized path name followed by a newline and exit
--syscall arg -- Given a number, print the matching syscall name and exit
                 Given a regex, print case-insensitively matching syscall names and exit
--sh          -- Output a shell script which defines \"esyd\" the SydB☮x helper function
--test        -- Run integration tests and exit
                 Requires syd-test & syd-test-do programs to be in PATH.

-M, --unshare-mount   -- Unshare mount namespace
-S, --unshare-uts     -- Unshare UTS namespace
-I, --unshare-ipc     -- Unshare IPC namespace
-U, --unshare-user    -- Unshare user namespace
-P, --unshare-pid     -- Unshare pid namespace
-N, --unshare-net     -- Unshare net namespace
-G, --unshare-cgroup  -- Unshare cgroup namespace

-A, --arg0 alias      -- Set alias of the command
                         Passed as `argv[0]` to the program
-0, --map-root        -- Map current user to root in the sandbox
                         Implies --unshare-user
-R, --chroot=root     -- Change directory and chroot to root before starting the process
                         The directory root/proc must exist to mount private procfs
                         Implies --unshare-mount and --unshare-pid

* syd exec may be used to construct a sandbox command to execute a
process outside the sandbox. See the description of cmd/exec
command in the README for more information.
* syd log may be used to access sandbox logs using journalctl.

# Profiles

1. container: Enables Linux namespaces.
              Equivalent to --unshare-mount,uts,ipc,user,pid,net,cgroup
2. landlock: Enables LandLock and allowlists system directories
             for Lock Sandboxing.
3. paludis: Used by the Paludis package mangler.
4. noipv4: Disables IPv4 connectivity.
5. noipv6: Disables Ipv6 connectivity.
6. silent: Silences all access violations.
7. lib: LibSyd helper profile. Turns all sandboxing off.
        Useful to configure SydB☮x in the application using LibSyd.
8. user: Allows user-specific directories and connections,
         and parses the following files if they exist:
         1. /etc/user.syd-3
         2. ~/.user.syd-3

When invoked without arguments, the current shell is executed
under sandbox with the user profile.

# Environment

SYD_LOG: Set log level. See the Enabling Logging¹ section of
 env-logger crate documentation for more information.
SYD_NO_SYSLOG: Disable logging to syslog(3). By default logs of
               severity Warn and higher are logged to syslog(3).
SYD_NO_CROSS_MEMORY_ATTACH: Disable cross memory attach,
                            and fallback to /proc/$pid/mem.
SYD_FAKEROOT: Equivalent to passing --root
SYD_SH: Pick the shell to spawn when invoked as a login shell
        Defaults to /bin/bash
SYD_UNSHARE_MOUNT: Equivalent to passing --unshare-mount
SYD_UNSHARE_UTS: Equivalent to passing --unshare-uts
SYD_UNSHARE_IPC: Equivalent to passing --unshare-ipc
SYD_UNSHARE_PID: Equivalent to passing --unshare-pid
SYD_UNSHARE_NET: Equivalent to passing --unshare-net
SYD_UNSHARE_CGROUP: Equivalent to passing --unshare-cgroup

# Exit Codes

Sydb☮x exits with the same exit code as the sandbox process
itself. If the sandbox process exits with a signal, Sydb☮x exits
with 14 which stands for EFAULT. In case there was an error in
spawning or waiting for the sandbox process, Sydb☮x exits with
errno indicating the error condition. E.g. syd true\" returns 0,
\"syd false\" return 1, and \"syd -- syd true\" returns 16 which
stands for EBUSY which stands for \"Device or resource busy\"
indicating there is already a secure computing filter
loaded. tl;dr Sydb☮x won't run under Sydb☮x, similarly many
process inspection tools such as \"ltrace\", \"strace\", or
\"gdb\" won't work under Sydb☮x. Thus the sandbox process can
either be traced by attaching from outside the sandbox or
running the tracer in follow fork mode, e.g.
\"strace -f syd true\".

¹: https://docs.rs/env_logger/latest/env_logger/#enabling-logging",
        *SYD_VERSION, libmaj, libmin, libmic, libapi, native
    );
}

/// Print version information.
fn version() {
    println!("SydB☮x {}", *SYD_VERSION);
    println!("Auth☮r: Ali P☮latel <alip@chessw☮b.☮rg>");
    println!("License: GPL-3.0-☮r-later");

    let feat = [
        #[cfg(debug_assertions)]
        "+debug",
        #[cfg(not(debug_assertions))]
        "-debug",
        #[cfg(feature = "log")]
        "+log",
        #[cfg(not(feature = "log"))]
        "-log",
        #[cfg(feature = "uring")]
        "+uring",
        #[cfg(not(feature = "uring"))]
        "-uring",
    ];
    println!("Features: {}", feat.join(", "));

    let state = match syd::lock_enabled() {
        0 => "Fully enf☮rced",
        1 => "Partially enf☮rced",
        2 => "Not enf☮rced",
        _ => "Unsupp☮rted",
    };
    println!("LandL☮ck: {state}");

    let libapi = libseccomp::get_api();
    match ScmpVersion::current() {
        Ok(libver) => {
            println!(
                "LibSecc☮mp: v{}.{}.{} api:{}",
                libver.major, libver.minor, libver.micro, libapi
            );
        }
        Err(error) => {
            println!("LibSecc☮mp: ? (error: {error})");
        }
    }

    let uname = match uname() {
        Ok(info) => OsStr::to_str(info.release()).unwrap_or("?").to_string(),
        Err(_) => "?".to_string(),
    };
    println!(
        "H☮st: {uname} {}",
        syd::seccomp_arch_native_name().unwrap_or("?")
    );
    println!("C☮mp: {}", env!("SYD_BUILDHOST"));
}

// Parse CLI Arguments
#[allow(clippy::cognitive_complexity)]
fn parse_argv() -> anyhow::Result<(String, Arguments)> {
    let mut args = Arguments::default();

    // Initialize namespace options from the environment.
    args.sandbox
        .set_unshare_mount(env::var(ENV_UNSHARE_MOUNT).is_ok());
    args.sandbox
        .set_unshare_uts(env::var(ENV_UNSHARE_UTS).is_ok());
    args.sandbox
        .set_unshare_ipc(env::var(ENV_UNSHARE_IPC).is_ok());
    args.sandbox
        .set_unshare_user(env::var(ENV_UNSHARE_USER).is_ok());
    args.sandbox
        .set_unshare_pid(env::var(ENV_UNSHARE_PID).is_ok());
    args.sandbox
        .set_unshare_net(env::var(ENV_UNSHARE_NET).is_ok());
    args.sandbox
        .set_unshare_cgroup(env::var(ENV_UNSHARE_CGROUP).is_ok());

    // Initialize Options.
    let mut argv: Vec<String> = env::args().collect();
    let is_login = argv.remove(0).starts_with('-');

    // We want to parse user profile early in two cases:
    // 1. No positional arguments were passed.
    // 2. The -c argument was passed.
    // This way the user can override it, e.g:
    // syd -psandbox/pid:off -c "echo hello"
    let parse_user_profile = if is_login {
        true
    } else {
        let mut has_opt = false;
        let mut has_pos = false;
        for arg in argv {
            if !arg.starts_with('-') {
                has_pos = true;
                break;
            } else if matches!(arg.as_str(), "-c" | "-l" | "--login") {
                has_opt = true;
                break;
            }
        }
        has_opt || !has_pos
    };
    if parse_user_profile {
        args.sandbox
            .parse_profile("user")
            .context("Error parsing user profile")?;
    }
    // Note: If someone ever needs/wants non-UTF-8 CLI arguments,
    // we can consider changing this, but till then we consider
    // this safe and sounds.
    #[allow(clippy::disallowed_methods)]
    let argv = argv::iter()
        .skip(1)
        .map(|os| os.to_str().expect("Argument is not valid UTF-8!"));
    let mut opts = Options::new(argv);

    // Local options handled by this function.
    let mut arch = ScmpArch::Native; // --arch
    let mut call = None; // --syscall
    let mut test = false; // --test
    let mut list_set: Option<String> = None; // --list
    let mut print_profile: Option<String> = None; // --print
    let sh = env::var(ENV_SH).unwrap_or(SYD_SH.to_string());

    while let Some(opt) = opts.next_opt().context("calling Options::next")? {
        match opt {
            /*
             * Basic options
             */
            Opt::Short('h') | Opt::Long("help") => {
                help();
                exit(0);
            }
            Opt::Short('v') | Opt::Long("version") => {
                version();
                exit(0);
            }

            /*
             * Sandbox options
             */
            Opt::Long("check") => {
                exit(if syd::syd_enabled() { 0 } else { 1 });
            }
            Opt::Long("landlock") => {
                exit(i32::from(syd::lock_enabled()));
            }
            Opt::Long("sh") => {
                println!("{ESYD}");
                exit(0);
            }
            Opt::Long("export") => {
                args.export = Some(ExportMode::from_str(
                    opts.value().context("--export requires an argument!")?,
                )?);
                args.sandbox
                    .parse_profile("paludis")
                    .context("Unable to parse paludis profile!")?;
            }
            Opt::Long("list") => {
                list_set = Some(
                    opts.value()
                        .context("--list requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("print") => {
                print_profile = Some(
                    opts.value()
                        .context("--print requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("errno") => {
                let val = opts.value().context("--errno requires an argument!")?;
                match val.parse::<u16>() {
                    Ok(0) => {
                        exit(1);
                    }
                    Ok(num) => {
                        // number -> name
                        let errno = Errno::from_i32(i32::from(num));
                        if errno == Errno::UnknownErrno {
                            exit(1);
                        }
                        let estr = errno.to_string();
                        let mut iter = estr.split(": ");
                        let name = iter.next().unwrap_or("?");
                        let desc = iter.next().unwrap_or("?");
                        println!("{num}\t{name}\t{desc}");
                    }
                    Err(_) => {
                        match regex::RegexBuilder::new(val)
                            .unicode(false)
                            .case_insensitive(true)
                            .build()
                        {
                            Ok(pattern) => {
                                // regex -> [number]
                                let mut ok = false;
                                for errno in (1..u8::MAX).map(|n| Errno::from_i32(i32::from(n))) {
                                    if errno == Errno::UnknownErrno {
                                        continue;
                                    }
                                    let estr = errno.to_string();
                                    let mut iter = estr.split(": ");
                                    let name = iter.next().unwrap_or("?");
                                    let desc = iter.next().unwrap_or("?");
                                    if pattern.is_match(name) {
                                        println!("{}\t{}\t{}", errno as i32, name, desc);
                                        ok = true;
                                    }
                                }
                                if !ok {
                                    exit(1);
                                }
                            }
                            Err(error) => {
                                eprintln!("Invalid errno regex \"{val}\": {error}");
                                exit(1);
                            }
                        }
                    }
                }
                exit(0);
            }
            Opt::Long("parse") => {
                let value = opts.value().context("--parse requires an argument!")?;
                if value.chars().all(|c| c.is_ascii_digit()) {
                    match value.parse::<usize>() {
                        Ok(size) => {
                            println!("{}", syd::human_size(size));
                            exit(0);
                        }
                        Err(error) => {
                            bail!("Failed to parse: {error}");
                        }
                    }
                } else {
                    match parse_size(value) {
                        Ok(size) => {
                            println!("{size}");
                            exit(0);
                        }
                        Err(error) => {
                            bail!("Failed to parse: {error}");
                        }
                    }
                }
            }
            Opt::Long("read") => {
                let path = Path::new(opts.value().context("--read requires an argument!")?);
                let path = if path.is_relative() {
                    Path::new("/proc/self/cwd").join(path)
                } else {
                    path.to_path_buf()
                };
                let path = syd::fs::canonicalize(path, true, syd::fs::MissingHandling::Missing)
                    .context("Failed to canonicalize path!")?;
                let path = path.as_os_str().as_bytes();
                std::io::stdout()
                    .write_all(path)
                    .context("Error writing canonicalized path to standard output!")?;
                println!();
                exit(0);
            }
            Opt::Long("arch") => {
                let val = opts.value()
                    .context("--arch requires an argument, do '--arch list' to print the list of architectures.")?;
                if matches!(val.to_ascii_lowercase().as_str(), "help" | "list") {
                    syd::print_seccomp_architectures();
                    exit(0);
                }
                arch = ScmpArch::from_str(&format!("SCMP_ARCH_{}", val.to_ascii_uppercase()))
                    .context(format!("Invalid architecture `{val}'"))?;
            }
            Opt::Long("syscall") => {
                call = Some(opts.value().context("--syscall requires an argument!")?);
            }
            Opt::Long("test") => {
                test = true;
            }
            Opt::Long("lock") => {
                args.sandbox.lock();
            }
            Opt::Long("root") => {
                args.sandbox.set_root(true);
            }
            Opt::Short('R') | Opt::Long("chroot") => {
                let val = opts.value().context("--chroot requires an argument!")?;
                if !val.starts_with('/') {
                    bail!("Chroot directory must be an absolute path!");
                }
                args.root = Some(PathBuf::from(val));
                args.sandbox.set_unshare_mount(true);
                args.sandbox.set_unshare_pid(true);
            }
            Opt::Short('m') | Opt::Long("magic") => {
                let cmd = opts.value().context("--magic requires an argument!")?;
                if args.sandbox.locked() {
                    error!("ctx": "config", "cfg": cmd, "err": "LOCK");
                } else {
                    args.sandbox
                        .config(cmd)
                        .context(format!("Failed to execute magic command `{cmd}'."))?;
                }
            }
            Opt::Short('l') | Opt::Long("login") => {
                // Ignored, login shell compatibility
            }
            Opt::Short('c') => {
                // When multiple -c arguments are given,
                // only the first one is honoured and
                // the rest is ignored in consistency
                // with how bash and dash behaves.
                if args.argv.is_empty() {
                    args.argv.push(sh.clone());
                    args.argv.push("-c".to_string());
                    args.argv.push(
                        opts.value()
                            .context("-c requires an argument!")?
                            .to_string(),
                    );
                }
            }
            Opt::Short('C') | Opt::Long("config") => {
                let path = opts.value().context("--config requires an argument!")?;
                if args.sandbox.locked() {
                    error!("ctx": "config", "path": path, "err": "LOCK");
                } else {
                    args.sandbox
                        .parse_config_file(path)
                        .context(format!("Failed to parse configuration file `{path}'."))?;
                }
            }
            Opt::Short('p') | Opt::Long("profile") => {
                let profile = opts.value().context("--profile requires an argument!")?;
                if args.sandbox.locked() {
                    error!("ctx": "config", "profile": profile, "err": "LOCK");
                } else {
                    args.sandbox.parse_profile(profile).context(format!(
                        "Failed to parse configuration profile `{profile}'."
                    ))?;
                }
            }

            /*
             * Unshare options
             */
            Opt::Short('M') | Opt::Long("unshare-mount") => {
                args.sandbox.set_unshare_mount(true);
            }
            Opt::Short('S') | Opt::Long("unshare-uts") => {
                args.sandbox.set_unshare_uts(true);
            }
            Opt::Short('I') | Opt::Long("unshare-ipc") => {
                args.sandbox.set_unshare_ipc(true);
            }
            Opt::Short('U') | Opt::Long("unshare-user") => {
                args.sandbox.set_unshare_user(true);
            }
            Opt::Short('P') | Opt::Long("unshare-pid") => {
                args.sandbox.set_unshare_pid(true);
            }
            Opt::Short('N') | Opt::Long("unshare-net") => {
                args.sandbox.set_unshare_net(true);
            }
            Opt::Short('G') | Opt::Long("unshare-cgroup") => {
                args.sandbox.set_unshare_cgroup(true);
            }
            Opt::Short('A') | Opt::Long("arg0") => {
                let val = opts
                    .value()
                    .context("--arg0 requires an argument!")?
                    .to_string();
                args.arg0 = Some(val);
            }
            Opt::Short('0') | Opt::Long("map-root") => {
                args.map_root = true;
                args.sandbox.set_unshare_user(true);
            }

            Opt::Short(c) => {
                bail!("Invalid option `-{c}'!");
            }
            Opt::Long(c) => {
                bail!("Invalid option `--{c}'!");
            }
        }
    }

    if test {
        Command::new("syd-test").args(opts.positionals()).exec();
    }

    // --syscall
    if let Some(val) = call {
        handle_opt_syscall(val, arch);
    }

    // --list and --print
    handle_opt_list_print(list_set.as_deref(), print_profile.as_deref());

    // Prepare the command to execute, which may be a login shell.
    args.argv.extend(opts.positionals().map(|s| s.to_string()));
    let argv0 = if !args.argv.is_empty() {
        Some(args.argv.remove(0))
    } else {
        None
    };

    let argv0 = match (args.export, argv0, is_login) {
        (Some(_), _, false) => "true".to_string(),
        (_, Some(cmd), false) if cmd == "log" => {
            Command::new("journalctl")
                .arg("SYSLOG_IDENTIFIER=syd")
                .exec();
            exit(127);
        }
        (_, Some(cmd), false) if cmd == "exec" => {
            // Split the arguments using the ASCII Unit Separator character
            let args = args.argv.into_iter().collect::<Vec<_>>().join("\x1F");

            // Format it using /dev/syd/cmd/exec!<concatenated-path>
            let path = format!("/dev/syd/cmd/exec!{args}");
            print!("{path}");
            exit(0);
        }
        (_, Some(argv0), false) => argv0.to_string(),
        (_, None, false) | (_, _, true) => {
            if args.arg0.is_none() {
                // Allow user to override with --arg0
                args.arg0 = Some("-".to_string());
            }
            sh.to_string()
        }
    };

    Ok((argv0, args))
}

/// Handle syd --syscall
fn handle_opt_syscall(val: &str, arch: ScmpArch) -> ! {
    match val.parse::<i32>() {
        Ok(num) => {
            let syscall = ScmpSyscall::from(num);
            if let Ok(name) = syscall.get_name_by_arch(arch) {
                println!("{num}\t{name}");
            } else {
                exit(1);
            }
        }
        Err(_) => {
            match regex::RegexBuilder::new(val)
                .unicode(false)
                .case_insensitive(true)
                .build()
            {
                Ok(pattern) => {
                    let mut ok = false;
                    for (num, name) in (0..4096)
                        .map(|n| {
                            (
                                n,
                                ScmpSyscall::from(n)
                                    .get_name_by_arch(arch)
                                    .unwrap_or_default(),
                            )
                        })
                        .filter(|(_, name)| !name.is_empty())
                    {
                        if pattern.is_match(&name) {
                            println!("{num}\t{name}");
                            ok = true;
                        }
                    }
                    if !ok {
                        exit(1);
                    }
                }
                Err(error) => {
                    eprintln!("Invalid syscall regex \"{val}\": {error}");
                    exit(1);
                }
            }
        }
    }
    exit(0);
}

/// Handle syd --list and syd --profile
#[allow(clippy::cognitive_complexity)]
fn handle_opt_list_print(list_set: Option<&str>, print_profile: Option<&str>) {
    // List set or Print profile if requested, and exit.
    match (list_set, print_profile) {
        (Some(_), Some(_)) => {
            eprintln!("The options --list and --print are mutually exclusive!");
            exit(1);
        }
        (Some("ioctl"), _) => {
            for (_, name) in ALLOWLIST_IOCTL {
                println!("{name}");
            }
            exit(0);
        }
        (Some("prctl"), _) => {
            for (_, name) in ALLOWLIST_PRCTL {
                println!("{name}");
            }
            exit(0);
        }
        (Some("allow"), _) => {
            for name in SAFE_SYSCALLS {
                println!("{name}");
            }
            exit(0);
        }
        (Some("hook"), _) => {
            for name in HOOK_SYSCALLS {
                println!("{name}");
            }
            exit(0);
        }
        (Some("deny"), _) => {
            let mut syscall_set: HashSet<_> =
                SAFE_SYSCALLS.iter().map(|&s| String::from(s)).collect();
            for syscall in HOOK_SYSCALLS {
                syscall_set.insert(syscall.to_string());
            }
            for syscall_number in 0..=600 {
                let syscall = ScmpSyscall::from(syscall_number);
                if let Ok(name) = syscall.get_name() {
                    if !syscall_set.contains(&name) {
                        println!("{name}");
                    }
                }
            }
            exit(0);
        }
        (Some(set), _) => {
            eprintln!("No such set: '{set}'");
            exit(0);
        }
        (_, Some("container")) => {
            println!("# SydB☮x profile: Container");
            println!("# Number of rules: {}", PROFILE_CONTAINER.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_CONTAINER {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("landlock")) => {
            println!("# SydB☮x profile: LandLock");
            println!("# Number of rules: {}", PROFILE_LANDLOCK.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_LANDLOCK {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("paludis")) => {
            println!("# SydB☮x profile: Paludis");
            println!("# Number of rules: {}", PROFILE_PALUDIS.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_PALUDIS {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("noipv4")) => {
            println!("# SydB☮x profile: NoIpv4");
            println!("# Number of rules: {}", PROFILE_NOIPV4.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV4 {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("noipv6")) => {
            println!("# SydB☮x profile: NoIpv6");
            println!("# Number of rules: {}", PROFILE_NOIPV6.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV6 {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("silent")) => {
            println!("# SydB☮x profile: Silent");
            println!("# Number of rules: {}", PROFILE_SILENT.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_SILENT {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("lib")) => {
            println!("# SydB☮x profile: LibSyd");
            println!("# Number of rules: {}", PROFILE_SILENT.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_LIB {
                println!("{command}");
            }
            exit(0);
        }
        (_, Some("user")) => {
            let uid = nix::unistd::getuid();
            let name = syd::get_user_name(uid);
            let mut home = syd::get_user_home(&name);

            println!("# SydB☮x profile: User \"{name}\"");
            println!(
                "# Number of rules: {}",
                PROFILE_USER.len().saturating_add(9)
            );
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");

            // Step 1: Static configuration defined at compile time.
            for command in PROFILE_USER {
                println!("{command}");
            }

            // Step 2: Dynamic, user-specific configuration.
            // Keep in sync with Sandbox::parse_profile()
            if !home.ends_with('/') {
                home.push('/');
            }

            println!("allowlist/lock/write+{home}");
            println!("allowlist/read+{home}***");
            println!("allowlist/stat+{home}***");
            println!("allowlist/write+{home}**");
            println!("allowlist/exec+{home}**");
            println!("allowlist/net/bind+{home}**");
            println!("allowlist/net/connect+{home}**");
            println!("allowlist/read+/run/user/{uid}/**");
            println!("allowlist/write+/run/user/{uid}/**");

            exit(0);
        }
        (_, Some(name)) => {
            eprintln!("No such profile: '{name}'");
            exit(1);
        }
        (None, None) => {
            // Do nothing, fall through.
        }
    };
}
